'''
File: soem.py
Project: link
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 07/11/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''

import ctypes
from ctypes import c_void_p, byref
from .link import Link

from pyautd3.native_methods.autd3capi_link_soem import NativeMethods as LinkSOEM


OnLostFunc = ctypes.CFUNCTYPE(None, ctypes.c_char_p)
LogOutputFunc = ctypes.CFUNCTYPE(None, ctypes.c_char_p)
LogFlushFunc = ctypes.CFUNCTYPE(None)


class SOEM:
    def __init__(self):
        self._ifname = None
        self._send_cycle = 2
        self._sync0_cycle = 2
        self._on_lost = None
        self._high_precision = False
        self._freerun = False
        self._check_interval = 500

    def ifname(self, ifname: str):
        self._ifname = ifname
        return self

    def send_cycle(self, cycle: int):
        self._send_cycle = cycle
        return self

    def sync0_cycle(self, cycle: int):
        self._sync0_cycle = cycle
        return self

    def on_lost(self, handle):
        self._on_lost = handle
        return self

    def high_precision(self, flag: bool):
        self._high_precision = flag
        return self

    def freerun(self, flag: bool):
        self._freerun = flag
        return self

    def check_interval(self, interval: int):
        self._check_interval = interval
        return self

    def build(self):
        LinkSOEM().init_dll()
        link = c_void_p()
        LinkSOEM().dll.AUTDLinkSOEM(byref(link), self._ifname.encode('utf-8') if self._ifname is not None else None,
                                    self._sync0_cycle, self._send_cycle, self._freerun, self._on_lost, self._high_precision, self._check_interval)
        return Link(link)

    @ staticmethod
    def enumerate_adapters():
        LinkSOEM().init_dll()
        res = []
        handle = c_void_p()
        size = LinkSOEM().dll.AUTDGetAdapterPointer(byref(handle))

        for i in range(size):
            sb_desc = ctypes.create_string_buffer(128)
            sb_name = ctypes.create_string_buffer(128)
            LinkSOEM().dll.AUTDGetAdapter(handle, i, sb_desc, sb_name)
            res.append([sb_name.value.decode('utf-8'), sb_desc.value.decode('utf-8')])

        LinkSOEM().dll.AUTDFreeAdapterPointer(handle)

        return res

    @ staticmethod
    def set_log_level(level: int):
        LinkSOEM().init_dll()
        LinkSOEM().dll.AUTDLinkSOEMSetLogLevel(level)

    @ staticmethod
    def set_log_func(output, flush):
        LinkSOEM().init_dll()
        LinkSOEM().dll.AUTDLinkSOEMSetDefaultLogger(output, flush)



'''
File: cuda_backend.py
Project: holo
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''

from ctypes import byref

from pyautd3.native_methods.autd3capi_backend_cuda import NativeMethods as BackendCUDA
from .backend import Backend


class CUDABackend(Backend):
    def __init__(self):
        super().__init__()
        BackendCUDA().init_dll()
        BackendCUDA().dll.AUTDCUDABackend(byref(self.ptr))

    def __del__(self):
        super().__del__()



'''
File: __init__.py
Project: gain
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from .primitive import Focus, BesselBeam, PlaneWave, Custom, Null, Grouped


__all__ = [
    'Focus',
    'BesselBeam',
    'PlaneWave',
    'Custom',
    'Null',
    'Grouped'
]



'''
File: __init__.py
Project: modulation
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from .primitive import Static, Custom, Sine, SineSquared, SineLegacy, Square
from .audio_file import RawPCM, Wav


__all__ = [
    'Static',
    'Custom',
    'Sine',
    'SineSquared',
    'SineLegacy',
    'Square',
    'RawPCM',
    'Wav',
]



'''
File: __init__.py
Project: holo
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 25/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''

from .backend import EigenBackend
from .cuda_backend import CUDABackend

from .constraint import Clamp, DontCare, Normalize, Uniform

from .apo import APO
from .evd import EVD
from .sdp import SDP
from .gs import GS
from .gspat import GSPAT
from .lm import LM
from .greedy import Greedy
from .lss_greedy import LSSGreedy
from .naive import Naive

__all__ = [
    'EigenBackend',
    'CUDABackend',
    'Clamp',
    'DontCare',
    'Normalize',
    'Uniform',
    'APO',
    'EVD',
    'SDP',
    'GS',
    'GSPAT',
    'LM',
    'Greedy',
    'LSSGreedy',
    'Naive'
]



'''
File: remote_twincat.py
Project: link
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 04/11/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from ctypes import c_void_p, byref

from .link import Link

from pyautd3.native_methods.autd3capi_link_remote_twincat import NativeMethods as LinkRemoteTwinCAT


class RemoteTwinCAT:
    def __init__(self, remote_ip_addr, remote_ams_net_id):
        self._remote_ip_addr = remote_ip_addr
        self._remote_ams_net_id = remote_ams_net_id
        self._local_ams_net_id = ''

    def local_ams_net_id(self, local_ams_net_id):
        self._local_ams_net_id = local_ams_net_id
        return self

    def build(self):
        link = c_void_p()
        LinkRemoteTwinCAT().dll.AUTDLinkRemoteTwinCAT(byref(link), self._remote_ip_addr.encode('utf-8'),
                                                      self._remote_ams_net_id.encode('utf-8'),
                                                      self._local_ams_net_id.encode('utf-8'))
        return Link(link)



'''
File: primitive.py
Project: gain
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''

from ctypes import byref
import numpy as np

from pyautd3.autd import Controller
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from .gain import Gain


class Focus(Gain):
    def __init__(self, pos, amp: float = 1.0):
        super().__init__()
        Base().dll.AUTDGainFocus(byref(self.ptr), pos[0], pos[1], pos[2], amp)

    def __del__(self):
        super().__del__()


class BesselBeam(Gain):
    def __init__(self, pos, dir, theta_z, amp: float = 1.0):
        super().__init__()
        Base().dll.AUTDGainBesselBeam(byref(self.ptr), pos[0], pos[1], pos[2], dir[0], dir[1], dir[2], theta_z, amp)

    def __del__(self):
        super().__del__()


class PlaneWave(Gain):
    def __init__(self, pos, dir, amp: float = 1.0):
        super().__init__()
        Base().dll.AUTDGainPlaneWave(byref(self.ptr), pos[0], pos[1], pos[2], dir[0], dir[1], dir[2], amp)

    def __del__(self):
        super().__del__()


class Custom(Gain):
    def __init__(self, amp, phase):
        super().__init__()
        p_size = len(phase)
        phase = np.ctypeslib.as_ctypes(np.array(phase).astype(np.double))
        amp = np.ctypeslib.as_ctypes(np.array(amp).astype(np.double))
        Base().dll.AUTDGainCustom(byref(self.ptr), amp, phase, p_size)

    def __del__(self):
        super().__del__()


class Null(Gain):
    def __init__(self):
        super().__init__()
        Base().dll.AUTDGainNull(byref(self.ptr))

    def __del__(self):
        super().__del__()


class Grouped(Gain):
    def __init__(self, autd: Controller):
        super().__init__()
        Base().dll.AUTDGainGrouped(byref(self.ptr), autd.p_cnt)

    def __del__(self):
        super().__del__()

    def add(self, dev_idx: int, gain: Gain):
        Base().dll.AUTDGainGroupedAdd(self.ptr, dev_idx, gain.ptr)



'''
File: __init__.py
Project: link
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 07/11/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''

from .remote_twincat import RemoteTwinCAT
from .soem import SOEM, OnLostFunc, LogOutputFunc, LogFlushFunc
from .twincat import TwinCAT
from .simulator import Simulator
from .remote_soem import RemoteSOEM

__all__ = [
    'RemoteTwinCAT',
    'SOEM',
    'OnLostFunc',
    'LogOutputFunc',
    'LogFlushFunc',
    'TwinCAT',
    'Simulator',
    'RemoteSOEM'
]



'''
File: constraint.py
Project: holo
Created Date: 25/10/2022
Author: Shun Suzuki
-----
Last Modified: 25/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from ctypes import c_void_p, byref

from pyautd3.native_methods.autd3capi_gain_holo import NativeMethods as GainHolo


class AmplitudeConstraint():
    def __init__(self):
        self.ptr = c_void_p()


class DontCare(AmplitudeConstraint):
    def __init__(self):
        super().__init__()
        GainHolo().dll.AUTDConstraintDontCare(byref(self.ptr))


class Normalize(AmplitudeConstraint):
    def __init__(self):
        super().__init__()
        GainHolo().dll.AUTDConstraintNormalize(byref(self.ptr))


class Uniform(AmplitudeConstraint):
    def __init__(self, value: float):
        super().__init__()
        GainHolo().dll.AUTDConstraintUniform(byref(self.ptr), value)


class Clamp(AmplitudeConstraint):
    def __init__(self):
        super().__init__()
        GainHolo().dll.AUTDConstraintClamp(byref(self.ptr))



'''
File: simulator.py
Project: extra
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 29/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from pyautd3.native_methods.autd3capi_extra_simulator import NativeMethods as ExtraSimulator


class Simulator:
    def __init__(self):
        self._settings_path = "settings.json"
        self._vsync = True
        self._gpu_idx = 0

    def settings_path(self, value: str):
        self._settings_path = value
        return self

    def vsync(self, value: bool):
        self._vsync = value
        return self

    def gpu_idx(self, value: int):
        self._gpu_idx = value
        return self

    def run(self):
        ExtraSimulator().init_dll()
        ExtraSimulator().dll.AUTDExtraSimulator(self._settings_path.encode('utf-8'), self._vsync, self._gpu_idx)



'''
File: evd.py
Project: holo
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from .holo import Holo
from .backend import Backend

from ctypes import byref

from pyautd3.native_methods.autd3capi_gain_holo import NativeMethods as GainHolo


class EVD(Holo):
    def __init__(self, backend: Backend, gamma: float = 1.0):
        super().__init__()
        GainHolo().dll.AUTDGainHoloEVD(byref(self.ptr), backend.ptr, gamma)

    def __del__(self):
        super().__del__()



'''
File: link.py
Project: link
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


class Link:
    def __init__(self, link):
        self.link_ptr = link



'''
File: primitive.py
Project: gain
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''

from ctypes import byref
import numpy as np

from pyautd3.native_methods.autd3capi import NativeMethods as Base
from .modulation import Modulation


class Static(Modulation):
    def __init__(self, amp: float = 1.0):
        super().__init__()
        Base().dll.AUTDModulationStatic(byref(self.ptr), amp)

    def __del__(self):
        super().__del__()


class Custom(Modulation):
    def __init__(self, data, sampling_freq_div: int):
        super().__init__()
        size = len(data)
        data = np.array(data).astype(np.uint8)
        data = np.ctypeslib.as_ctypes(data)

        Base().dll.AUTDModulationCustom(byref(self.ptr), data, size, sampling_freq_div)

    def __del__(self):
        super().__del__()


class Sine(Modulation):
    def __init__(self, freq: int, amp: float = 1.0, offset: float = 0.5):
        super().__init__()
        Base().dll.AUTDModulationSine(byref(self.ptr), freq, amp, offset)

    def __del__(self):
        super().__del__()


class SineSquared(Modulation):
    def __init__(self, freq: int, amp: float = 1.0, offset: float = 0.5):
        super().__init__()
        Base().dll.AUTDModulationSineSquared(byref(self.ptr), freq, amp, offset)

    def __del__(self):
        super().__del__()


class SineLegacy(Modulation):
    def __init__(self, freq: float, amp: float = 1.0, offset: float = 0.5):
        super().__init__()
        Base().dll.AUTDModulationSineLegacy(byref(self.ptr), freq, amp, offset)

    def __del__(self):
        super().__del__()


class Square(Modulation):
    def __init__(self, freq: int, low: float = 0.0, high: float = 1.0, duty: float = 0.5):
        super().__init__()
        Base().dll.AUTDModulationSquare(byref(self.ptr), freq, low, high, duty)

    def __del__(self):
        super().__del__()



'''
File: holo.py
Project: holo
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 25/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''

from pyautd3.native_methods.autd3capi_gain_holo import NativeMethods as GainHolo
from pyautd3.gain.gain import Gain
from .constraint import AmplitudeConstraint


class Holo(Gain):
    def __init__(self):
        super().__init__()

    def __del__(self):
        super().__del__()

    def add(self, focus, amp):
        GainHolo().dll.AUTDGainHoloAdd(self.ptr, focus[0], focus[1], focus[2], amp)

    def amplitude_constraint(self, constraint: AmplitudeConstraint):
        GainHolo().dll.AUTDSetConstraint(self.ptr, constraint.ptr)



'''
File: __init__.py
Project: pyautd3
Created Date: 24/05/2021
Author: Shun Suzuki
-----
Last Modified: 16/12/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from pyautd3.autd import SilencerConfig
from pyautd3.autd import Controller
from pyautd3.autd import Amplitudes
from pyautd3.autd import NUM_TRANS_IN_UNIT, NUM_TRANS_X, NUM_TRANS_Y, TRANS_SPACING, DEVICE_HEIGHT, DEVICE_WIDTH
from pyautd3.autd import DRIVER_LATEST, DRIVER_V2_2, DRIVER_V2_3, DRIVER_V2_4, DRIVER_V2_5, DRIVER_V2_6
from pyautd3.autd import Clear, UpdateFlag, Synchronize, ModDelayConfig, Stop
from pyautd3.autd import LogOutputFunc, LogFlushFunc, set_log_level, set_log_func

__all__ = [
    'SilencerConfig',
    'Controller',
    'Amplitudes',
    'Clear',
    'UpdateFlag',
    'Synchronize',
    'ModDelayConfig',
    'Stop',
    'NUM_TRANS_IN_UNIT',
    'NUM_TRANS_X',
    'NUM_TRANS_Y',
    'TRANS_SPACING',
    'DEVICE_WIDTH',
    'DEVICE_HEIGHT',
    'DRIVER_LATEST',
    'DRIVER_V2_2',
    'DRIVER_V2_3',
    'DRIVER_V2_4',
    'DRIVER_V2_5',
    'DRIVER_V2_6',
    'LogOutputFunc',
    'LogFlushFunc',
    'set_log_level',
    'set_log_func',
]

__version__ = '2.7.0'



# This file was automatically generated from header file
import threading
import ctypes
import os


class Singleton(type):
    _instances = {}
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):
    def init_path(self, bin_location: str, bin_prefix: str, bin_ext: str):
        self.bin = bin_location
        self.prefix = bin_prefix
        self.ext = bin_ext

    def init_dll(self):
        if hasattr(self, 'dll'):
            return
        self.dll = ctypes.CDLL(os.path.join(self.bin, f'{self.prefix}autd3capi{self.ext}'))

        self.dll.AUTDSetLogLevel.argtypes = [ctypes.c_int32] 
        self.dll.AUTDSetLogLevel.restype = None

        self.dll.AUTDSetDefaultLogger.argtypes = [ctypes.c_void_p, ctypes.c_void_p] 
        self.dll.AUTDSetDefaultLogger.restype = None

        self.dll.AUTDCreateController.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_uint8] 
        self.dll.AUTDCreateController.restype = ctypes.c_bool

        self.dll.AUTDOpenController.argtypes = [ctypes.c_void_p, ctypes.c_void_p] 
        self.dll.AUTDOpenController.restype = ctypes.c_bool

        self.dll.AUTDAddDevice.argtypes = [ctypes.c_void_p, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double] 
        self.dll.AUTDAddDevice.restype = None

        self.dll.AUTDAddDeviceQuaternion.argtypes = [ctypes.c_void_p, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double] 
        self.dll.AUTDAddDeviceQuaternion.restype = None

        self.dll.AUTDClose.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDClose.restype = ctypes.c_bool

        self.dll.AUTDFreeController.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDFreeController.restype = None

        self.dll.AUTDIsOpen.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDIsOpen.restype = ctypes.c_bool

        self.dll.AUTDGetForceFan.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDGetForceFan.restype = ctypes.c_bool

        self.dll.AUTDGetReadsFPGAInfo.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDGetReadsFPGAInfo.restype = ctypes.c_bool

        self.dll.AUTDGetAckCheckTimeout.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDGetAckCheckTimeout.restype = ctypes.c_uint64

        self.dll.AUTDGetSendInterval.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDGetSendInterval.restype = ctypes.c_uint64

        self.dll.AUTDSetReadsFPGAInfo.argtypes = [ctypes.c_void_p, ctypes.c_bool] 
        self.dll.AUTDSetReadsFPGAInfo.restype = None

        self.dll.AUTDSetAckCheckTimeout.argtypes = [ctypes.c_void_p, ctypes.c_uint64] 
        self.dll.AUTDSetAckCheckTimeout.restype = None

        self.dll.AUTDSetSendInterval.argtypes = [ctypes.c_void_p, ctypes.c_uint64] 
        self.dll.AUTDSetSendInterval.restype = None

        self.dll.AUTDSetForceFan.argtypes = [ctypes.c_void_p, ctypes.c_bool] 
        self.dll.AUTDSetForceFan.restype = None

        self.dll.AUTDGetSoundSpeed.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDGetSoundSpeed.restype = ctypes.c_double

        self.dll.AUTDSetSoundSpeed.argtypes = [ctypes.c_void_p, ctypes.c_double] 
        self.dll.AUTDSetSoundSpeed.restype = None

        self.dll.AUTDSetSoundSpeedFromTemp.argtypes = [ctypes.c_void_p, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double] 
        self.dll.AUTDSetSoundSpeedFromTemp.restype = None

        self.dll.AUTDGetTransFrequency.argtypes = [ctypes.c_void_p, ctypes.c_int32] 
        self.dll.AUTDGetTransFrequency.restype = ctypes.c_double

        self.dll.AUTDSetTransFrequency.argtypes = [ctypes.c_void_p, ctypes.c_int32, ctypes.c_double] 
        self.dll.AUTDSetTransFrequency.restype = None

        self.dll.AUTDGetTransCycle.argtypes = [ctypes.c_void_p, ctypes.c_int32] 
        self.dll.AUTDGetTransCycle.restype = ctypes.c_uint16

        self.dll.AUTDSetTransCycle.argtypes = [ctypes.c_void_p, ctypes.c_int32, ctypes.c_uint16] 
        self.dll.AUTDSetTransCycle.restype = None

        self.dll.AUTDGetWavelength.argtypes = [ctypes.c_void_p, ctypes.c_int32] 
        self.dll.AUTDGetWavelength.restype = ctypes.c_double

        self.dll.AUTDGetAttenuation.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDGetAttenuation.restype = ctypes.c_double

        self.dll.AUTDSetAttenuation.argtypes = [ctypes.c_void_p, ctypes.c_double] 
        self.dll.AUTDSetAttenuation.restype = None

        self.dll.AUTDGetFPGAInfo.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_uint8)] 
        self.dll.AUTDGetFPGAInfo.restype = ctypes.c_bool

        self.dll.AUTDNumTransducers.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDNumTransducers.restype = ctypes.c_int32

        self.dll.AUTDNumDevices.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDNumDevices.restype = ctypes.c_int32

        self.dll.AUTDGeometryCenter.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double)] 
        self.dll.AUTDGeometryCenter.restype = None

        self.dll.AUTDTransPosition.argtypes = [ctypes.c_void_p, ctypes.c_int32, ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double)] 
        self.dll.AUTDTransPosition.restype = None

        self.dll.AUTDTransXDirection.argtypes = [ctypes.c_void_p, ctypes.c_int32, ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double)] 
        self.dll.AUTDTransXDirection.restype = None

        self.dll.AUTDTransYDirection.argtypes = [ctypes.c_void_p, ctypes.c_int32, ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double)] 
        self.dll.AUTDTransYDirection.restype = None

        self.dll.AUTDTransZDirection.argtypes = [ctypes.c_void_p, ctypes.c_int32, ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double)] 
        self.dll.AUTDTransZDirection.restype = None

        self.dll.AUTDGetFirmwareInfoListPointer.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_void_p)] 
        self.dll.AUTDGetFirmwareInfoListPointer.restype = ctypes.c_int32

        self.dll.AUTDGetFirmwareInfo.argtypes = [ctypes.c_void_p, ctypes.c_int32, ctypes.c_char_p] 
        self.dll.AUTDGetFirmwareInfo.restype = None

        self.dll.AUTDFreeFirmwareInfoListPointer.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDFreeFirmwareInfoListPointer.restype = None

        self.dll.AUTDGainNull.argtypes = [ctypes.POINTER(ctypes.c_void_p)] 
        self.dll.AUTDGainNull.restype = None

        self.dll.AUTDGainGrouped.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_void_p] 
        self.dll.AUTDGainGrouped.restype = None

        self.dll.AUTDGainGroupedAdd.argtypes = [ctypes.c_void_p, ctypes.c_int32, ctypes.c_void_p] 
        self.dll.AUTDGainGroupedAdd.restype = None

        self.dll.AUTDGainFocus.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double] 
        self.dll.AUTDGainFocus.restype = None

        self.dll.AUTDGainBesselBeam.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double] 
        self.dll.AUTDGainBesselBeam.restype = None

        self.dll.AUTDGainPlaneWave.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double] 
        self.dll.AUTDGainPlaneWave.restype = None

        self.dll.AUTDGainTransducerTest.argtypes = [ctypes.POINTER(ctypes.c_void_p)] 
        self.dll.AUTDGainTransducerTest.restype = None

        self.dll.AUTDGainTransducerTestSet.argtypes = [ctypes.c_void_p, ctypes.c_int32, ctypes.c_double, ctypes.c_double] 
        self.dll.AUTDGainTransducerTestSet.restype = None

        self.dll.AUTDGainCustom.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.c_uint64] 
        self.dll.AUTDGainCustom.restype = None

        self.dll.AUTDDeleteGain.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDDeleteGain.restype = None

        self.dll.AUTDModulationStatic.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_double] 
        self.dll.AUTDModulationStatic.restype = None

        self.dll.AUTDModulationSine.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_int32, ctypes.c_double, ctypes.c_double] 
        self.dll.AUTDModulationSine.restype = None

        self.dll.AUTDModulationSineSquared.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_int32, ctypes.c_double, ctypes.c_double] 
        self.dll.AUTDModulationSineSquared.restype = None

        self.dll.AUTDModulationSineLegacy.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_double, ctypes.c_double, ctypes.c_double] 
        self.dll.AUTDModulationSineLegacy.restype = None

        self.dll.AUTDModulationSquare.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_int32, ctypes.c_double, ctypes.c_double, ctypes.c_double] 
        self.dll.AUTDModulationSquare.restype = None

        self.dll.AUTDModulationLPF.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_void_p] 
        self.dll.AUTDModulationLPF.restype = None

        self.dll.AUTDModulationCustom.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint64, ctypes.c_uint32] 
        self.dll.AUTDModulationCustom.restype = None

        self.dll.AUTDModulationSamplingFrequencyDivision.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDModulationSamplingFrequencyDivision.restype = ctypes.c_uint32

        self.dll.AUTDModulationSetSamplingFrequencyDivision.argtypes = [ctypes.c_void_p, ctypes.c_uint32] 
        self.dll.AUTDModulationSetSamplingFrequencyDivision.restype = None

        self.dll.AUTDModulationSamplingFrequency.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDModulationSamplingFrequency.restype = ctypes.c_double

        self.dll.AUTDDeleteModulation.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDDeleteModulation.restype = None

        self.dll.AUTDFocusSTM.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_double] 
        self.dll.AUTDFocusSTM.restype = None

        self.dll.AUTDGainSTM.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_void_p] 
        self.dll.AUTDGainSTM.restype = None

        self.dll.AUTDFocusSTMAdd.argtypes = [ctypes.c_void_p, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_uint8] 
        self.dll.AUTDFocusSTMAdd.restype = None

        self.dll.AUTDGainSTMAdd.argtypes = [ctypes.c_void_p, ctypes.c_void_p] 
        self.dll.AUTDGainSTMAdd.restype = None

        self.dll.AUTDGetGainSTMMode.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDGetGainSTMMode.restype = ctypes.c_uint16

        self.dll.AUTDSetGainSTMMode.argtypes = [ctypes.c_void_p, ctypes.c_uint16] 
        self.dll.AUTDSetGainSTMMode.restype = None

        self.dll.AUTDSTMSetFrequency.argtypes = [ctypes.c_void_p, ctypes.c_double] 
        self.dll.AUTDSTMSetFrequency.restype = ctypes.c_double

        self.dll.AUTDSTMGetStartIdx.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDSTMGetStartIdx.restype = ctypes.c_int32

        self.dll.AUTDSTMGetFinishIdx.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDSTMGetFinishIdx.restype = ctypes.c_int32

        self.dll.AUTDSTMSetStartIdx.argtypes = [ctypes.c_void_p, ctypes.c_int32] 
        self.dll.AUTDSTMSetStartIdx.restype = None

        self.dll.AUTDSTMSetFinishIdx.argtypes = [ctypes.c_void_p, ctypes.c_int32] 
        self.dll.AUTDSTMSetFinishIdx.restype = None

        self.dll.AUTDSTMFrequency.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDSTMFrequency.restype = ctypes.c_double

        self.dll.AUTDSTMSamplingFrequency.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDSTMSamplingFrequency.restype = ctypes.c_double

        self.dll.AUTDSTMSamplingFrequencyDivision.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDSTMSamplingFrequencyDivision.restype = ctypes.c_uint32

        self.dll.AUTDSTMSetSamplingFrequencyDivision.argtypes = [ctypes.c_void_p, ctypes.c_uint32] 
        self.dll.AUTDSTMSetSamplingFrequencyDivision.restype = None

        self.dll.AUTDDeleteSTM.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDDeleteSTM.restype = None

        self.dll.AUTDSynchronize.argtypes = [ctypes.POINTER(ctypes.c_void_p)] 
        self.dll.AUTDSynchronize.restype = None

        self.dll.AUTDClear.argtypes = [ctypes.POINTER(ctypes.c_void_p)] 
        self.dll.AUTDClear.restype = None

        self.dll.AUTDUpdateFlags.argtypes = [ctypes.POINTER(ctypes.c_void_p)] 
        self.dll.AUTDUpdateFlags.restype = None

        self.dll.AUTDStop.argtypes = [ctypes.POINTER(ctypes.c_void_p)] 
        self.dll.AUTDStop.restype = None

        self.dll.AUTDModDelayConfig.argtypes = [ctypes.POINTER(ctypes.c_void_p)] 
        self.dll.AUTDModDelayConfig.restype = None

        self.dll.AUTDDeleteSpecialData.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDDeleteSpecialData.restype = None

        self.dll.AUTDCreateSilencer.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_uint16, ctypes.c_uint16] 
        self.dll.AUTDCreateSilencer.restype = None

        self.dll.AUTDDeleteSilencer.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDDeleteSilencer.restype = None

        self.dll.AUTDSend.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p] 
        self.dll.AUTDSend.restype = ctypes.c_bool

        self.dll.AUTDSendSpecial.argtypes = [ctypes.c_void_p, ctypes.c_void_p] 
        self.dll.AUTDSendSpecial.restype = ctypes.c_bool

        self.dll.AUTDSendAsync.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p] 
        self.dll.AUTDSendAsync.restype = None

        self.dll.AUTDSendSpecialAsync.argtypes = [ctypes.c_void_p, ctypes.c_void_p] 
        self.dll.AUTDSendSpecialAsync.restype = None

        self.dll.AUTDGetModDelay.argtypes = [ctypes.c_void_p, ctypes.c_int32] 
        self.dll.AUTDGetModDelay.restype = ctypes.c_uint16

        self.dll.AUTDSetModDelay.argtypes = [ctypes.c_void_p, ctypes.c_int32, ctypes.c_uint16] 
        self.dll.AUTDSetModDelay.restype = None

        self.dll.AUTDCreateAmplitudes.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_double] 
        self.dll.AUTDCreateAmplitudes.restype = None

        self.dll.AUTDDeleteAmplitudes.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDDeleteAmplitudes.restype = None

        self.dll.AUTDSetMode.argtypes = [ctypes.c_void_p, ctypes.c_uint8] 
        self.dll.AUTDSetMode.restype = None

        self.dll.AUTDSoftwareSTM.argtypes = [ctypes.POINTER(ctypes.c_void_p)] 
        self.dll.AUTDSoftwareSTM.restype = None

        self.dll.AUTDSoftwareSTMSetStrategy.argtypes = [ctypes.c_void_p, ctypes.c_uint8] 
        self.dll.AUTDSoftwareSTMSetStrategy.restype = None

        self.dll.AUTDSoftwareSTMAdd.argtypes = [ctypes.c_void_p, ctypes.c_void_p] 
        self.dll.AUTDSoftwareSTMAdd.restype = None

        self.dll.AUTDSoftwareSTMStart.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_void_p, ctypes.c_void_p] 
        self.dll.AUTDSoftwareSTMStart.restype = None

        self.dll.AUTDSoftwareSTMFinish.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDSoftwareSTMFinish.restype = None

        self.dll.AUTDSoftwareSTMSetFrequency.argtypes = [ctypes.c_void_p, ctypes.c_double] 
        self.dll.AUTDSoftwareSTMSetFrequency.restype = ctypes.c_double

        self.dll.AUTDSoftwareSTMFrequency.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDSoftwareSTMFrequency.restype = ctypes.c_double

        self.dll.AUTDSoftwareSTMPeriod.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDSoftwareSTMPeriod.restype = ctypes.c_uint64

        self.dll.AUTDSoftwareSTMSamplingFrequency.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDSoftwareSTMSamplingFrequency.restype = ctypes.c_double

        self.dll.AUTDSoftwareSTMSamplingPeriod.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDSoftwareSTMSamplingPeriod.restype = ctypes.c_uint64

        self.dll.AUTDSoftwareSTMSetSamplingPeriod.argtypes = [ctypes.c_void_p, ctypes.c_uint64] 
        self.dll.AUTDSoftwareSTMSetSamplingPeriod.restype = None

        self.dll.AUTDDeleteSoftwareSTM.argtypes = [ctypes.c_void_p] 
        self.dll.AUTDDeleteSoftwareSTM.restype = None



'''
File: simulator.py
Project: link
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 29/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from ctypes import c_void_p, byref

from .link import Link

from pyautd3.native_methods.autd3capi_link_simulator import NativeMethods as LinkSimulator


class Simulator:
    def __init__(self):
        pass

    def build(self):
        link = c_void_p()
        LinkSimulator().dll.AUTDLinkSimulator(byref(link))
        return Link(link)



'''
File: backend.py
Project: holo
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from ctypes import c_void_p, byref

from pyautd3.native_methods.autd3capi_gain_holo import NativeMethods as GainHolo


class Backend():
    def __init__(self):
        self.ptr = c_void_p()

    def __del__(self):
        GainHolo().dll.AUTDDeleteBackend(self.ptr)


class EigenBackend(Backend):
    def __init__(self):
        super().__init__()
        GainHolo().dll.AUTDEigenBackend(byref(self.ptr))

    def __del__(self):
        super().__del__()



'''
File: greedy.py
Project: holo
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from .holo import Holo
from .backend import Backend

from ctypes import byref

from pyautd3.native_methods.autd3capi_gain_holo import NativeMethods as GainHolo


class Greedy(Holo):
    def __init__(self, backend: Backend, phase_div: int = 16):
        super().__init__()
        GainHolo().dll.AUTDGainHoloGreedy(
            byref(self.ptr),
            backend.ptr,
            phase_div)

    def __del__(self):
        super().__del__()



'''
File: modulation.py
Project: modulation
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.autd import Header


class Modulation(Header):
    def __init__(self):
        super().__init__()

    def __del__(self):
        Base().dll.AUTDDeleteModulation(self.ptr)

    @ property
    def sampling_frequency_division(self):
        return Base().dll.AUTDModulationSamplingFrequencyDivision(self.ptr)

    @ sampling_frequency_division.setter
    def sampling_frequency_division(self, value: int):
        return Base().dll.AUTDModulationSetSamplingFrequencyDivision(self.ptr, value)

    @ property
    def sampling_frequency(self):
        return Base().dll.AUTDModulationSamplingFrequency(self.ptr)



'''
File: remote_twincat.py
Project: link
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 04/11/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from ctypes import c_void_p, byref

from .link import Link

from pyautd3.native_methods.autd3capi_link_remote_soem import NativeMethods as LinkRemoteSOEM


class RemoteSOEM:
    def __init__(self):
        self._ip = ""
        self._port = 50632

    def ip(self, ip):
        self._ip = ip
        return self

    def port(self, port):
        self._port = port
        return self

    def build(self):
        link = c_void_p()
        LinkRemoteSOEM().dll.AUTDLinkRemoteSOEM(byref(link), self._ip.encode('utf-8'), self._port)
        return Link(link)



'''
File: __ini__.py
Project: stm
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 29/11/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''

from .stm import FocusSTM, GainSTM, Mode

__all__ = [
    'FocusSTM',
    'GainSTM',
    'Mode'
]



'''
File: stm.py
Project: stm
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 16/12/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''

from ctypes import byref
from enum import IntEnum

from pyautd3.autd import Body, Controller
from pyautd3.gain.gain import Gain
from pyautd3.native_methods.autd3capi import NativeMethods as Base


class STM(Body):
    def __init__(self):
        super().__init__()

    def __del__(self):
        Base().dll.AUTDDeleteSTM(self.ptr)

    @ property
    def frequency(self):
        return Base().dll.AUTDSTMFrequency(self.ptr)

    @ frequency.setter
    def frequency(self, freq: float):
        return Base().dll.AUTDSTMSetFrequency(self.ptr, freq)

    @ property
    def start_idx(self):
        return Base().dll.AUTDSTMGetStartIdx(self.ptr)

    @ start_idx.setter
    def start_idx(self, value: int):
        return Base().dll.AUTDSTMSetStartIdx(self.ptr, value)

    @ property
    def finish_idx(self):
        return Base().dll.AUTDSTMGetFinishIdx(self.ptr)

    @ finish_idx.setter
    def finish_idx(self, value: int):
        return Base().dll.AUTDSTMSetFinishIdx(self.ptr, value)

    @ property
    def sampling_frequency(self):
        return Base().dll.AUTDSTMSamplingFrequency(self.ptr)

    @ property
    def sampling_frequency_division(self):
        return Base().dll.AUTDSTMSamplingFrequencyDivision(self.ptr)

    @ sampling_frequency_division.setter
    def sampling_frequency_division(self, value: int):
        return Base().dll.AUTDSTMSetSamplingFrequencyDivision(self.ptr, value)


class FocusSTM(STM):
    def __init__(self, sound_speed: float):
        super().__init__()
        Base().dll.AUTDFocusSTM(byref(self.ptr), sound_speed)

    def __del__(self):
        super().__del__()

    def add(self, point, duty_shift: int = 0):
        Base().dll.AUTDFocusSTMAdd(self.ptr, point[0], point[1], point[2], duty_shift)


class Mode(IntEnum):
    PhaseDutyFull = 0x01
    PhaseFull = 0x02
    PhaseHalf = 0x04


class GainSTM(STM):
    def __init__(self, autd: Controller):
        super().__init__()
        Base().dll.AUTDGainSTM(byref(self.ptr), autd.p_cnt)

    def __del__(self):
        super().__del__()

    def add(self, gain: Gain):
        Base().dll.AUTDGainSTMAdd(self.ptr, gain.ptr)

    @ property
    def mode(self):
        return Mode(Base().dll.AUTDGetGainSTMMode(self.ptr))

    @ mode.setter
    def mode(self, value: Mode):
        Base().dll.AUTDSetGainSTMMode(self.ptr, int(value))



'''
File: __init__.py
Project: native_methods
Created Date: 09/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''

import os
import os.path
import platform

from .autd3capi import NativeMethods as Base
from .autd3capi_gain_holo import NativeMethods as GainHolo
from .autd3capi_modulation_audio_file import NativeMethods as ModulationAudioFile
from .autd3capi_backend_cuda import NativeMethods as BackendCUDA
from .autd3capi_link_simulator import NativeMethods as LinkSimulator
from .autd3capi_link_remote_twincat import NativeMethods as LinkRemoteTwinCAT
from .autd3capi_link_twincat import NativeMethods as LinkTwincAT
from .autd3capi_link_soem import NativeMethods as LinkSOEM
from .autd3capi_extra_simulator import NativeMethods as ExtraSimulator
from .autd3capi_extra_geometry_viewer import NativeMethods as ExtraGeometryViewer

_PLATFORM = platform.system()
_TARGET_OS = ''
_ARCH = ''
_PREFIX = ''
_BIN_EXT = ''
if _PLATFORM == 'Windows':
    _BIN_EXT = '.dll'
    _TARGET_OS = 'win'
    _ARCH = 'x64' if platform.machine().endswith('64') else 'x86'
elif _PLATFORM == 'Darwin':
    _PREFIX = 'lib'
    _BIN_EXT = '.dylib'
    _TARGET_OS = 'macos'
    _ARCH = 'universal'
elif _PLATFORM == 'Linux':
    _PREFIX = 'lib'
    _BIN_EXT = '.so'
    _TARGET_OS = 'linux'
    if platform.machine().startswith('aarch64'):
        _ARCH = 'arm64'
    elif platform.machine().startswith('arm64'):
        _ARCH = 'arm64'
    elif platform.machine().startswith('arm'):
        _ARCH = 'arm32'
    elif platform.machine().endswith('64'):
        _ARCH = 'x64'
    else:
        raise ImportError('Cannot identify CPU architecture')
else:
    raise ImportError('Not supported OS')

_LIB_PATH = os.path.join(os.path.dirname(__file__), '..', 'bin', f'{_TARGET_OS}_{_ARCH}')

Base().init_path(_LIB_PATH, _PREFIX, _BIN_EXT)
GainHolo().init_path(_LIB_PATH, _PREFIX, _BIN_EXT)
ModulationAudioFile().init_path(_LIB_PATH, _PREFIX, _BIN_EXT)
BackendCUDA().init_path(_LIB_PATH, _PREFIX, _BIN_EXT)
Base().init_dll()
GainHolo().init_dll()
ModulationAudioFile().init_dll()

LinkRemoteTwinCAT().init_path(_LIB_PATH, _PREFIX, _BIN_EXT)
LinkSimulator().init_path(_LIB_PATH, _PREFIX, _BIN_EXT)
LinkSOEM().init_path(_LIB_PATH, _PREFIX, _BIN_EXT)
LinkTwincAT().init_path(_LIB_PATH, _PREFIX, _BIN_EXT)
LinkRemoteTwinCAT().init_dll()
LinkSimulator().init_dll()
if platform.system() == 'Windows':
    LinkTwincAT().init_dll()

ExtraSimulator().init_path(_LIB_PATH, _PREFIX, _BIN_EXT)
ExtraGeometryViewer().init_path(_LIB_PATH, _PREFIX, _BIN_EXT)



'''
File: __init__.py
Project: extra
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from .geometry_viewer import GeometryViewer
from .simulator import Simulator


__all__ = [
    'GeometryViewer',
    'Simulator'
]



'''
File: gain.py
Project: gain
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''

from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.autd import Body


class Gain(Body):
    def __init__(self):
        super().__init__()

    def __del__(self):
        Base().dll.AUTDDeleteGain(self.ptr)



'''
File: autd.py
Project: pyautd3
Created Date: 24/05/2021
Author: Shun Suzuki
-----
Last Modified: 09/12/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


import ctypes
from ctypes import c_void_p, byref, c_double
import numpy as np

from .native_methods.autd3capi import NativeMethods as Base
from .link.link import Link


NUM_TRANS_IN_UNIT = 249
NUM_TRANS_X = 18
NUM_TRANS_Y = 14
TRANS_SPACING = 10.16
DEVICE_WIDTH = 192.0
DEVICE_HEIGHT = 151.4

DRIVER_LATEST = 0x00
DRIVER_V2_2 = 0x82
DRIVER_V2_3 = 0x83
DRIVER_V2_4 = 0x84
DRIVER_V2_5 = 0x85
DRIVER_V2_6 = 0x86


LogOutputFunc = ctypes.CFUNCTYPE(None, ctypes.c_char_p)
LogFlushFunc = ctypes.CFUNCTYPE(None)


def set_log_level(level: int):
    Base().dll.AUTDSetLogLevel(level)


def set_log_func(output, flush):
    Base().dll.AUTDSetDefaultLogger(output, flush)


class SpecialData:
    def __init__(self):
        self.ptr = c_void_p()


class Body:
    def __init__(self):
        self.ptr = c_void_p()


class Header:
    def __init__(self):
        self.ptr = c_void_p()


class SilencerConfig(Header):
    def __init__(self, step: int = 10, cycle: int = 4096):
        super().__init__()
        Base().dll.AUTDCreateSilencer(byref(self.ptr), step, cycle)

    def __del__(self):
        Base().dll.AUTDDeleteSilencer(self.ptr)

    @staticmethod
    def none():
        return SilencerConfig(0xFFFF, 4096)


class Transducer:
    def __init__(self, tr_id: int, cnt: c_void_p):
        self._tr_id = tr_id
        self._cnt = cnt

    @ property
    def id(self) -> int:
        return self._tr_id

    @ property
    def position(self):
        x = c_double(0.0)
        y = c_double(0.0)
        z = c_double(0.0)
        Base().dll.AUTDTransPosition(self._cnt, self._tr_id, byref(x), byref(y), byref(z))
        return np.array([x.value, y.value, z.value])

    @ property
    def frequency(self):
        return Base().dll.AUTDGetTransFrequency(self._cnt, self._tr_id)

    @ frequency.setter
    def frequency(self, freq: float):
        return Base().dll.AUTDSetTransFrequency(self._cnt, self._tr_id, freq)

    @ property
    def cycle(self):
        return Base().dll.AUTDGetTransCycle(self._cnt, self._tr_id)

    @ cycle.setter
    def cycle(self, cycle: int):
        return Base().dll.AUTDSetTransCycle(self._cnt, self._tr_id, cycle)

    @ property
    def mod_delay(self):
        return Base().dll.AUTDGetModDelay(self._cnt, self._tr_id)

    @ mod_delay.setter
    def mod_delay(self, delay: int):
        return Base().dll.AUTDSetModDelay(self._cnt, self._tr_id, delay)

    @ property
    def wavelength(self):
        return Base().dll.AUTDGetWavelength(self._cnt, self._tr_id)

    @ property
    def x_direction(self):
        x = c_double(0.0)
        y = c_double(0.0)
        z = c_double(0.0)
        Base().dll.AUTDTransXDirection(self._cnt, self._tr_id, byref(x), byref(y), byref(z))
        return np.array([x.value, y.value, z.value])

    @ property
    def y_direction(self):
        x = c_double(0.0)
        y = c_double(0.0)
        z = c_double(0.0)
        Base().dll.AUTDTransYDirection(self._cnt, self._tr_id, byref(x), byref(y), byref(z))
        return np.array([x.value, y.value, z.value])

    @ property
    def z_direction(self):
        x = c_double(0.0)
        y = c_double(0.0)
        z = c_double(0.0)
        Base().dll.AUTDTransZDirection(self._cnt, self._tr_id, byref(x), byref(y), byref(z))
        return np.array([x.value, y.value, z.value])


class Geometry:
    def __init__(self, cnt: c_void_p):
        self._cnt = cnt

    def add_device(self, pos, rot):
        Base().dll.AUTDAddDevice(self._cnt, pos[0], pos[1], pos[2], rot[0], rot[1], rot[2])

    def add_device_quaternion(self, pos, q):
        Base().dll.AUTDAddDeviceQuaternion(self._cnt, pos[0], pos[1], pos[2], q[0], q[1], q[2], q[3])

    @ property
    def num_transducers(self):
        return Base().dll.AUTDNumTransducers(self._cnt)

    @ property
    def num_devices(self):
        return Base().dll.AUTDNumDevices(self._cnt)

    @property
    def center(self):
        x = c_double(0.0)
        y = c_double(0.0)
        z = c_double(0.0)
        Base().dll.AUTDGeometryCenter(self._cnt, byref(x), byref(y), byref(z))
        return np.array([x.value, y.value, z.value])

    def __getitem__(self, key: int):
        return Transducer(key, self._cnt)

    class TransdducerIterator:
        def __init__(self, cnt: c_void_p):
            self._cnt = cnt
            self._i = 0
            self._num_trans = Base().dll.AUTDNumTransducers(cnt)

        def __next__(self):
            if self._i == self._num_trans:
                raise StopIteration()
            value = Transducer(self._i, self._cnt)
            self._i += 1
            return value

    def __iter__(self):
        return Geometry.TransdducerIterator(self._cnt)


class Controller:
    def __init__(self, driver_version: int = DRIVER_LATEST):
        self.p_cnt = c_void_p()
        Base().dll.AUTDCreateController(byref(self.p_cnt), driver_version)
        self.__disposed = False

    def __del__(self):
        self.dispose()

    def to_legacy(self):
        Base().dll.AUTDSetMode(self.p_cnt, 0)

    def to_normal(self):
        Base().dll.AUTDSetMode(self.p_cnt, 1)

    def to_normal_phaseself(self):
        Base().dll.AUTDSetMode(self.p_cnt, 2)

    @ property
    def geometry(self):
        return Geometry(self.p_cnt)

    @ property
    def sound_speed(self):
        return Base().dll.AUTDGetSoundSpeed(self.p_cnt)

    @ sound_speed.setter
    def sound_speed(self, sound_speed: float):
        Base().dll.AUTDSetSoundSpeed(self.p_cnt, sound_speed)

    def set_sound_speed_from_temp(self, temp: float, k: float = 1.4, r: float = 8.31446261815324, m: float = 28.9647e-3):
        Base().dll.AUTDSetSoundSpeed(self.p_cnt, temp, k, r, m)

    @ property
    def attenuation(self):
        return Base().dll.AUTDGetAttenuation(self.p_cnt)

    @ attenuation.setter
    def attenuation(self, attenuation: float):
        Base().dll.AUTDSetAttenuation(self.p_cnt, attenuation)

    def open(self, link: Link):
        return Base().dll.AUTDOpenController(self.p_cnt, link.link_ptr)

    def firmware_info_list(self):
        res = []
        handle = c_void_p()
        size = Base().dll.AUTDGetFirmwareInfoListPointer(self.p_cnt, byref(handle))

        for i in range(size):
            sb = ctypes.create_string_buffer(256)
            Base().dll.AUTDGetFirmwareInfo(handle, i, sb)
            res.append(sb.value.decode('utf-8'))

        Base().dll.AUTDFreeFirmwareInfoListPointer(handle)

        return res

    def dispose(self):
        if not self.__disposed:
            self.close()
            self._free()
            self.__disposed = True

    def close(self):
        return Base().dll.AUTDClose(self.p_cnt)

    def _free(self):
        Base().dll.AUTDFreeController(self.p_cnt)

    @ property
    def is_open(self):
        return Base().dll.AUTDIsOpen(self.p_cnt)

    @ property
    def force_fan(self):
        return Base().dll.AUTDGetForceFan(self.p_cnt)

    @ force_fan.setter
    def force_fan(self, value: bool):
        return Base().dll.AUTDSetForceFan(self.p_cnt, value)

    @ property
    def ack_check_timeout_ms(self):
        return Base().dll.AUTDGetAckCheckTimeout(self.p_cnt) / 1000 / 1000

    @ ack_check_timeout_ms.setter
    def ack_check_timeout_ms(self, value: int):
        return Base().dll.AUTDSetAckCheckTimeout(self.p_cnt, value * 1000 * 1000)

    @ property
    def ack_check_timeout_ns(self):
        return Base().dll.AUTDGetAckCheckTimeout(self.p_cnt)

    @ ack_check_timeout_ns.setter
    def ack_check_timeout_ns(self, value: int):
        return Base().dll.AUTDSetAckCheckTimeout(self.p_cnt, value)

    @ property
    def send_interval_ms(self):
        return Base().dll.AUTDGetSendInterval(self.p_cnt) / 1000 / 1000

    @ send_interval_ms.setter
    def send_interval_ms(self, value: int):
        return Base().dll.AUTDSetSendInterval(self.p_cnt, value * 1000 * 1000)

    @ property
    def send_interval_ns(self):
        return Base().dll.AUTDGetSendInterval(self.p_cnt)

    @ send_interval_ns.setter
    def send_interval_ns(self, value: int):
        return Base().dll.AUTDSetSendInterval(self.p_cnt, value)

    @ property
    def reads_fpga_info(self):
        Base().dll.AUTDGetReadsFPGAInfo(self.p_cnt)

    @ reads_fpga_info.setter
    def reads_fpga_info(self, value: bool):
        Base().dll.AUTDSetReadsFPGAInfo(self.p_cnt, value)

    @ property
    def fpga_info(self):
        infos = np.zeros([self.num_transducers() / NUM_TRANS_IN_UNIT]).astype(np.ubyte)
        pinfos = np.ctypeslib.as_ctypes(infos)
        Base().dll.AUTDGetFPGAInfo(self.p_cnt, pinfos)
        return infos

    def send(self, a, b=None):
        if b is None and isinstance(a, SpecialData):
            return Base().dll.AUTDSendSpecial(self.p_cnt, a.ptr)
        if b is None and isinstance(a, Header):
            return Base().dll.AUTDSend(self.p_cnt, a.ptr, None)
        if b is None and isinstance(a, Body):
            return Base().dll.AUTDSend(self.p_cnt, None, a.ptr)
        if isinstance(a, Header) and isinstance(b, Body):
            return Base().dll.AUTDSend(self.p_cnt, a.ptr, b.ptr)
        raise NotImplementedError()

    def send_async(self, a, b=None):
        if b is None and isinstance(a, SpecialData):
            Base().dll.AUTDSendSpecialAsync(self.p_cnt, a.ptr)
            a.ptr = c_void_p()
            return
        if b is None and isinstance(a, Header):
            Base().dll.AUTDSendAsync(self.p_cnt, a.ptr, None)
            a.ptr = c_void_p()
            return
        if b is None and isinstance(a, Body):
            Base().dll.AUTDSendAsync(self.p_cnt, None, a.ptr)
            a.ptr = c_void_p()
            return
        if isinstance(a, Header) and isinstance(b, Body):
            Base().dll.AUTDSendAsync(self.p_cnt, a.ptr, b.ptr)
            a.ptr = c_void_p()
            b.ptr = c_void_p()
            return
        raise NotImplementedError()


class Amplitudes(Body):
    def __init__(self, amp: float):
        super().__init__()
        Base().dll.AUTDCreateAmplitudes(byref(self.ptr), amp)

    def __del__(self):
        Base().dll.AUTDDeleteAmplitudes(self.ptr)


class Clear(SpecialData):
    def __init__(self):
        super().__init__()
        Base().dll.AUTDClear(byref(self.ptr))

    def __del__(self):
        Base().dll.AUTDDeleteSpecialData(self.ptr)


class Stop(SpecialData):
    def __init__(self):
        super().__init__()
        Base().dll.AUTDStop(byref(self.ptr))

    def __del__(self):
        Base().dll.AUTDDeleteSpecialData(self.ptr)


class UpdateFlag(SpecialData):
    def __init__(self):
        super().__init__()
        Base().dll.AUTDUpdateFlags(byref(self.ptr))

    def __del__(self):
        Base().dll.AUTDDeleteSpecialData(self.ptr)


class Synchronize(SpecialData):
    def __init__(self):
        super().__init__()
        Base().dll.AUTDSynchronize(byref(self.ptr))

    def __del__(self):
        Base().dll.AUTDDeleteSpecialData(self.ptr)


class ModDelayConfig(SpecialData):
    def __init__(self):
        super().__init__()
        Base().dll.AUTDModDelayConfig(byref(self.ptr))

    def __del__(self):
        Base().dll.AUTDDeleteSpecialData(self.ptr)



'''
File: twincat.py
Project: link
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''

from ctypes import c_void_p, byref

from .link import Link

from pyautd3.native_methods.autd3capi_link_twincat import NativeMethods as LinkTwinCAT


class TwinCAT:
    def __init__(self):
        pass

    def build(self):
        link = c_void_p()
        LinkTwinCAT().dll.AUTDLinkTwinCAT(byref(link))
        return Link(link)



'''
File: geometry_viewer.py
Project: extra
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from pyautd3.autd import Geometry
from pyautd3.native_methods.autd3capi_extra_geometry_viewer import NativeMethods as ExtraGeometryViewer


class GeometryViewer:
    def __init__(self):
        self._width = 800
        self._height = 600
        self._vsync = True
        self._gpu_idx = 0

    def window_size(self, width: int, height: int):
        self._width = width
        self._height = height
        return self

    def vsync(self, value: bool):
        self._vsync = value
        return self

    def gpu_idx(self, value: int):
        self._gpu_idx = value
        return self

    def view(self, geometry: Geometry):
        ExtraGeometryViewer().init_dll()
        ExtraGeometryViewer().dll.AUTDExtraGeometryViewer(geometry._cnt, self._width, self._height, self._vsync, self._gpu_idx)



'''
File: setup.py
Project: pyautd
Created Date: 10/09/2020
Author: Shun Suzuki
-----
Last Modified: 25/11/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2021 Shun Suzuki. All rights reserved.
'''

import glob
import setuptools
import requests
import shutil
import zipfile
import tarfile
import os
import sys


def _get_version():
    with open('README.md', 'r') as f:
        for line in f.readlines():
            if line.startswith('version: '):
                return line.replace('version: ', '').strip()
    raise LookupError('version info is not found in README.md')


def _set_package_version(version):
    init_py = ''
    with open('pyautd3/__init__.py', 'r') as f:
        for line in f.readlines():
            if line.startswith('__version__'):
                line = '__version__ = \'' + version.strip() + '\'\n'
            init_py = init_py + line

    with open('pyautd3/__init__.py', 'w') as f:
        f.write(init_py)


with open('README.md', 'r') as fh:
    long_description = fh.read()

_set_package_version(_get_version())

_VERSION_TRIPLE = '.'.join(_get_version().split('.')[0:3])


class Target:
    bin_ext = ''
    archive_ext = ''
    os_name = ''
    arch = 'x64'

    def __init__(self, os_name, arch, bin_ext, archive_ext):
        self.os_name = os_name
        self.arch = arch
        self.bin_ext = bin_ext
        self.archive_ext = archive_ext


def download_bin():
    asset_base_url = 'https://github.com/shinolab/autd3/releases/download/'
    version = f'v{_VERSION_TRIPLE}'

    targets = [
        Target('win', 'x64', 'dll', 'zip'),
        Target('macos', 'universal', 'dylib', 'tar.gz'),
        Target('linux', 'x64', 'so', 'tar.gz'),
        Target('linux', 'arm64', 'so', 'tar.gz'),
        Target('linux', 'arm32', 'so', 'tar.gz'),
    ]

    for target in targets:
        url = f'{asset_base_url}{version}/autd3-{version}-{target.os_name}-{target.arch}.{target.archive_ext}'

        tmp_archive_path = f'tmp.{target.archive_ext}'

        res = requests.get(url, stream=True)
        with open(tmp_archive_path, 'wb') as fp:
            shutil.copyfileobj(res.raw, fp)

        if target.archive_ext == 'zip':
            with zipfile.ZipFile(tmp_archive_path) as f:
                for info in f.infolist():
                    if info.filename.startswith('bin') and info.filename.endswith(target.bin_ext):
                        f.extract(info, '.')
        elif target.archive_ext == 'tar.gz':
            with tarfile.open(tmp_archive_path) as f:
                libraries = []
                for i in f.getmembers():
                    if i.name.startswith('bin') and i.name.endswith(target.bin_ext):
                        libraries.append(i)
                f.extractall(path='.', members=libraries)

        os.makedirs(f'pyautd3/bin/{target.os_name}_{target.arch}', exist_ok=True)
        for f in glob.glob('./bin/*'):
            shutil.move(f, f'pyautd3/bin/{target.os_name}_{target.arch}')
        os.remove(tmp_archive_path)


skip_download_bin = False
if len(sys.argv) > 3:
    skip_download_bin = sys.argv[3] == 'True'
    del sys.argv[3]

if not skip_download_bin:
    download_bin()

packages = ['pyautd3',
            'pyautd3.native_methods',
            'pyautd3.gain',
            'pyautd3.gain.holo',
            'pyautd3.modulation',
            'pyautd3.link',
            'pyautd3.stm',
            'pyautd3.extra']
for f in glob.glob('pyautd3/bin/*'):
    f = f.replace('\\', '.').replace('/', '.')
    packages.append(f)

data_files = []
for f in glob.glob('pyautd3/bin/**/*'):
    data_files.append(f)

setuptools.setup(
    name='pyautd3',
    version=_get_version(),
    author='Shun Suzuki',
    author_email='suzuki@hapis.k.u-tokyo.ac.jp',
    description='AUTD3 library wrapper for python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/shinolab/autd3',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: MacOS'
    ],
    license='See https://github.com/shinolab/autd3/blob/master/LICENSE',
    platforms=["Windows", "Linux", "Mac OS-X"],
    include_package_data=True,
    package_dir={'pyautd3': 'pyautd3'},
    packages=packages,
    python_requires='>=3.8',
    data_files=[(
        'lib/site-packages', data_files
    )],
)



# This file was automatically generated from header file
import threading
import ctypes
import os


class Singleton(type):
    _instances = {}
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):
    def init_path(self, bin_location: str, bin_prefix: str, bin_ext: str):
        self.bin = bin_location
        self.prefix = bin_prefix
        self.ext = bin_ext

    def init_dll(self):
        if hasattr(self, 'dll'):
            return
        self.dll = ctypes.CDLL(os.path.join(self.bin, f'{self.prefix}autd3capi-link-remote-soem{self.ext}'))

        self.dll.AUTDLinkRemoteSOEM.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_char_p, ctypes.c_uint16] 
        self.dll.AUTDLinkRemoteSOEM.restype = None



'''
File: audio_file.py
Project: modulation
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''

from ctypes import byref

from pyautd3.native_methods.autd3capi_modulation_audio_file import NativeMethods as ModulationAudioFile

from .modulation import Modulation


class RawPCM(Modulation):
    def __init__(self, path: str, sampling_freq: float, mod_freq_div: int):
        super().__init__()
        ModulationAudioFile().dll.AUTDModulationRawPCM(byref(self.ptr), path.encode('utf-8'), sampling_freq, mod_freq_div)

    def __del__(self):
        super().__del__()


class Wav(Modulation):
    def __init__(self, path: str, mod_freq_div: int):
        super().__init__()
        ModulationAudioFile().dll.AUTDModulationWav(byref(self.ptr), path.encode('utf-8'), mod_freq_div)

    def __del__(self):
        super().__del__()



'''
File: lm.py
Project: holo
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 21/10/2022
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022 Shun Suzuki. All rights reserved.

'''


from .holo import Holo
from .backend import Backend

from ctypes import byref

from pyautd3.native_methods.autd3capi_gain_holo import NativeMethods as GainHolo


class LM(Holo):
    def __init__(self, backend: Backend, eps1: float = 1e-8, eps2: float = 1e-8, tau: float = 1e-3,
                 k_max: int = 5, initial=None):
        super().__init__()
        GainHolo().dll.AUTDGainHoloLM(
            byref(self.ptr),
            backend.ptr,
            eps1,
            eps2,
            tau,
            k_max,
            initial,
            0 if initial is None else len(initial))

    def __del__(self):
        super().__del__()
